#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include "WTlib.h"
#include "wavelet_v7.h"
#include "myallocs.h"

t_WaveletFamily **pWF = NULL;
unsigned int     LWF = 0;
int          errorWF = 0;

int Ipsi (double out, t_WaveletFamily *WF, double dt);
int WT2ST (t_CWTvar *x, t_WaveletFamily *WF, double dt);
int ST2WT (t_CWTvar *x, t_WaveletFamily *WF, double dt);

int create_wf_ (unsigned int *WFid, int *type, unsigned int *J, unsigned int *V, unsigned int *N, double *s0, double *b0, int *convtype, double *w0, int *continuous) {
	unsigned int id;
	
	/* Initializations. */
	*WFid   = 0;
	errorWF = 0;
	LWF++;
	id = LWF;
	
	/* printf("create_wf:%d  J:%d,  V:%d  N:%d  s0:%f  b0:%f  convtype:%d  w0:%f  continuous:%d \n", *type, *J, *V, *N, *s0, *b0, *convtype, *w0, *continuous); */
	
	/* Save memory to keep a pointer to the new WaveletFamily structure (WF). */
	if (pWF == NULL) pWF = (t_WaveletFamily **)mymalloc(sizeof(t_WaveletFamily *));
	else  pWF = (t_WaveletFamily **)myrealloc(pWF, LWF*sizeof(t_WaveletFamily *));
	
	/* If memory is properly saved, create the new WF. */
	if (pWF == NULL) errorWF = 1;     /* Error: Memory allocation problem for **pWF. */
	else {
		pWF[id-1] = CreateWaveletFamily (*type, *J, *V, *N, *s0, *b0, *convtype, *w0, *continuous);
		if (pWF == NULL) errorWF = 2; /* Error: WF creation. Allocation/wrong param. */
		else *WFid = id;              /* Everything is ok.                           */
	}
	
	return errorWF;
}

int destroy_wf_ (unsigned int *WFid) {
	if (*WFid > LWF) return 1;                      /* Wrong WFid.    */
	else DestroyWaveletFamily (pWF[*WFid-1]);       /* Remove the WF. */
	
	/* Released the memory unused (still, only the last elements.)   */
	if (*WFid == LWF) {
		for (; LWF>=0 && pWF[LWF]==NULL; LWF--);   
		if (LWF == 0) myfree(pWF);
		else pWF = (t_WaveletFamily **)myrealloc(pWF, LWF*sizeof(t_WaveletFamily *));
	}
	return 0;
}

int fwts_ (double complex *WTx, float *x, unsigned int *N, unsigned int *Nmax, unsigned int *WFid, int STlike, double *dt) {
	unsigned int n;
	int error;
	double *xdb;
	
	/* printf("fwts: %d  WFid: %d\n", *N, *WFid); */
	xdb = (double *)mymalloc(*N * sizeof(double));
	if (xdb == NULL) return 1;
	
	for (n=0; n<*N; n++) xdb[n] = (double)x[n];
	error = fwt_ (WTx, xdb, N, Nmax, WFid, STlike, dt);
	
	myfree(xdb);
	return error;
}

int ifwts_ (float *xrec, double complex *WTx, unsigned int *Ns, unsigned int *N, unsigned int *Nmax, unsigned int *WFid, int STlike, double *dt) {
	unsigned int n;
	int error;
	double *xdb;
	
	/* printf("Ns: %d  N: %d  WFid: %d \n", *Ns, *N, *WFid); */
	xdb = (double *)mymalloc(*N * sizeof(double));
	if (xdb == NULL) {
		printf("ifwts: xdb = NULL\n");
		return 1;
	}
	
	error = ifwt_ (xdb, WTx, Ns, N, Nmax, WFid, STlike, dt);
	for (n=0; n<*N; n++) xrec[n] = (float)xdb[n];
	if (error) printf("ifwts: error %d\n", error);
	
	myfree(xdb);
	return error;
}

int fwt_ (double complex *WTx, double *x, unsigned int *N, unsigned int *Nmax, unsigned int *WFid, int STlike, double *dt) {
	t_CWTvar *WTvar;
	unsigned int s, id;
	int error;
	double complex *pc;
	
	id = *WFid-1;
	WTvar = CreateComplexWaveletVar (pWF[id], *N);
	error = complex_1D_wavelet_dec (WTvar, x, *N, pWF[id]);	 /* The main job. */
	if(STlike) WT2ST(WTvar, pWF[id], *dt);
	
	/* Convert t_CWTvar structure to a matrix form. */
	pc = WTx;
	for (s=0; s<WTvar->S; s++) {
		memcpy(pc, WTvar->d[s], WTvar->N[s]*sizeof(double complex));
		pc += *Nmax;
	}

	DestroyComplexWaveletVar (WTvar);
	
	return error;
}

int ifwt_ (double *xrec, double complex *WTx, unsigned int *Ns, unsigned int *N, unsigned int *Nmax, unsigned int *WFid, int STlike, double *dt) {
	t_CWTvar *WTvar;
	unsigned int s, id;
	int error;
	double complex *pc;

	id = *WFid-1;
	if (pWF[id]->Ns > *Ns) {
		printf("Too many scales.");
		return -1;  /* Number of scales over the limit. */
	}
	
	/* Convert WTx to a t_CWTvar structure form. */
	WTvar = CreateComplexWaveletVar (pWF[id], *N);
	pc = WTx;
	for (s=0; s<WTvar->S; s++) {
		memcpy(WTvar->d[s], pc, WTvar->N[s]*sizeof(double complex));
        pc += *Nmax;
	}
	
	if(STlike) ST2WT(WTvar, pWF[id], *dt);
	error = Re_complex_1D_wavelet_rec (xrec, WTvar, *N, pWF[id]); /* The main job. */
	DestroyComplexWaveletVar (WTvar);
	
	return error;
}

int WT2ST (t_CWTvar *x, t_WaveletFamily *WF, double dt) {
	double complex *pc;
	double da1, da2, da3;
	unsigned int n, s;

	da1 = dt*WF->op1/(2*PI);
	if (!Ipsi(da1, WF, dt)) return 1;
	for (s=0; s<WF->Ns; s++) {
		pc  = x->d[s];
		da2 = da1/sqrt(WF->scale[s]);
		da3 = 2*PI*da1/WF->scale[s]*WF->Down_smp[s];
		for (n=0; n<WF->Ls[s]; n++)
			pc[n] *= da2 * cexp(I*da3*n);
	}
	return 0;
}

int ST2WT (t_CWTvar *x, t_WaveletFamily *WF, double dt) {
	complex *pc;
	double da1, da2, da3;
	unsigned int n, s;

	da1 = dt*WF->op1/(2*PI);
	if (!Ipsi(da1, WF, dt)) return 1;
	for (s=0; s<WF->Ns; s++) {
		pc  = x->d[s];
		da2 = sqrt(WF->scale[s])/da1;
		da3 = -2*PI*da1/WF->scale[s]*WF->Down_smp[s];
		for (n=0; n<WF->Ls[s]; n++)
			pc[n] *= da2 * cexp(I*da3*n);
	}
	return 0;
}

int Ipsi (double out, t_WaveletFamily *WF, double dt) {
	double Ipsi=0;
	double complex *pc;
	unsigned int Ns, n;
	
	Ns = WF->Ns;
	if (WF->format == 0) return 1;
	pc = WF->wframe.wc[Ns-1];
	for (n=0; n<WF->Ls[Ns-1]; n++)
		Ipsi += creal(pc[n])*creal(pc[n]) + cimag(pc[n])*cimag(pc[n]);
	Ipsi /= sqrt(WF->scale[Ns-1]);
	Ipsi *= dt;
	return 0;
}
